// Main
//------------------------------------------------------------------------------

// Includes
//------------------------------------------------------------------------------
#include "Worker/Worker.h"
#include "Core/Env/Assert.h"
#include "Core/Env/Env.h"
#include "Core/FileIO/FileIO.h"
#include "Core/Mem/MemTracker.h"
#include "Core/Process/Process.h"
#include "Core/Process/SystemMutex.h"
#include "Core/Process/Thread.h"
#include "Core/Strings/AStackString.h"

#include <windows.h>

// Global Data
//------------------------------------------------------------------------------
// only allow 1 worker per system
SystemMutex g_OneProcessMutex( "Global\\FBuildWorker" );

// Functions
//------------------------------------------------------------------------------
int LaunchSubProcess( const AString & args );

// main
//------------------------------------------------------------------------------
#pragma warning ( push )
#pragma warning (disable:28251) // don't complain about missing annotations on WinMain
int WINAPI WinMain( HINSTANCE hInstance, 
					HINSTANCE /*hPrevInstance*/,
					LPSTR lpCmdLine, 
					int /*nCmdShow*/ )
{
	// process cmd line args
	const AStackString<> args( lpCmdLine );
	Array< AString > tokens;
	args.Tokenize( tokens );
	bool isSubprocess = false; // are we the subprocess?
	bool useSubprocess = true; // do we want to launch as a sub-process?
	auto end = tokens.End();
	for ( auto it=tokens.Begin(); it != end; ++it )
	{
		const AString & token = *it;
		if ( token == "-subprocess" )
		{
			isSubprocess = true;
		}
		else if ( token == "-nosubprocess" )
		{
			useSubprocess = false;
		}
	}

	// only allow 1 worker per system
	Timer t;
	while ( g_OneProcessMutex.TryLock() == false )
	{
		// retry for upto 2 seconds, to allow some time for old worker to close
		if ( t.GetElapsed() > 5.0f )
		{
			MessageBoxA( nullptr, "An FBuildWorker is already running!", "FBuildWorker", MB_OK );
			return -1;
		}
		Thread::Sleep(100);
	}

	if ( useSubprocess && !isSubprocess )
	{
		return LaunchSubProcess( args );
	}

	// prevent popups when launching tools with missing dlls
	::SetErrorMode( SEM_FAILCRITICALERRORS );

	VERIFY( SetPriorityClass( GetCurrentProcess(), BELOW_NORMAL_PRIORITY_CLASS ) );

	// start the worker and wait for it to be closed
	int ret;
	{
		Worker worker( hInstance, args );
		ret = worker.Work();
	}

	MEMTRACKER_DUMP_ALLOCATIONS

	return ret;
}
#pragma warning ( pop )

// LaunchSubProcess
//------------------------------------------------------------------------------
int LaunchSubProcess( const AString & args )
{
	// try to make a copy of our exe
	AStackString<> exeName;
	Env::GetExePath( exeName );
	AStackString<> exeNameCopy( exeName );
	exeNameCopy += ".copy";
	Timer t;
	while ( FileIO::FileCopy( exeName.Get(), exeNameCopy.Get() ) == false )
	{
		if ( t.GetElapsed() > 5.0f )
		{
			MessageBoxA( nullptr, "Failed to make sub-process copy.", "FBuildWorker", MB_OK );
			return -2;
		}
		Thread::Sleep( 100 );
	}

	AStackString<> argsCopy( args );
	argsCopy += " -subprocess";

	// allow subprocess to access the mutex
	g_OneProcessMutex.Unlock();

	Process p;
	p.Spawn( exeNameCopy.Get(), argsCopy.Get(), nullptr, nullptr );
	p.Detach();

	return 0;
}

//------------------------------------------------------------------------------
